/*
 * *********************************************************************************************************************
 *
 *          !!
 *        .F88X
 *        X8888Y
 *      .}888888N;
 *        i888888N;        .:!              .I$WI:
 *          R888888I      .'N88~            i8}+8Y&8"l8i$8>8W~'>W8}8]KW+8IIN"8&
 *          .R888888I    .;N8888~          .X8'  "8I.!,/8"  !%NY8`"8I8~~8>,88I
 *            +888888N;  .8888888Y                                  "&&8Y.}8,
 *            ./888888N;  .R888888Y        .'}~    .>}'.`+>  i}!    "i'  +/'  .'i~  !11,.:">,  .~]!  .i}i
 *              ~888888%:  .I888888l      .]88~`1/iY88Ii+1'.R$8$8]"888888888>  Y8$  W8E  X8E  W8888'188Il}Y88$*
 *              18888888    E8888881    .]W%8$`R8X'&8%++N8i,8N%N8+l8%`  .}8N:.R$RE%N88N%N$K$R  188,FE$8%~Y88I
 *            .E888888I  .i8888888'      .:$8I;88+`E8R:/8N,.>881.`$8E/1/]N8X.Y8N`"KF&&FK!'88*."88K./$88%RN888+~
 *            8888888I  .,N888888~        ~88i"8W,!N8*.I88.}888%F,i$88"F88"  888:E8X.>88!i88>`888*.}Fl1]*}1YKi'
 *          i888888N'      I888Y          ]88;/EX*IFKFK88X  K8R  .l8W  88Y  ~88}'88E&%8W.X8N``]88!.$8K  .:W8I
 *        .i888888N;        I8Y          .&8$  .X88!  i881.:%888>I88  ;88]  +88+.';;;;:.Y88X  18N.,88l  .+88/
 *      .:R888888I
 *      .&888888I                                          Copyright (c) 2009-2020.  博锐尚格科技股份有限公司
 *        ~8888'
 *        .!88~                                                                     All rights reserved.
 *
 * *********************************************************************************************************************
 */

package com.persagy.database

import com.persagy.base.exceptions.SFieldNotFoundException
import com.persagy.base.extensions.getAllDeclaredField
import com.persagy.base.opts.SFormat
import com.persagy.database.enums.SDirection
import com.persagy.database.parser.SSyntaxParser
import java.io.ByteArrayInputStream
import java.text.SimpleDateFormat
import java.util.*
import kotlin.collections.ArrayList

/**
 * 查询条件生成器
 *
 * @author  庞利祥 <sybotan@126.com>
 */
class SQueryBuilder<ENTITY: Any>(
        private val baseDao: SAbstractDao<ENTITY>,
        val filterList: ArrayList<SFilter>) {

    init {
        // 转换数据类型
        for (item in filterList) {
            convertType(baseDao.entityClass, item)
        }
    } // Init

    /** 表名 */
    var tableName: String? = null
    /** 映射字段列表 */
    var fieldList = ArrayList<String>()
    /** 排序条件 */
    var orderArgs = ArrayList<SOrder>()
    /** 是否去除重复记录 */
    var isDistinct = false
    /** 偏移量 */
    var offset = 0
        set(value) {
            field = value
            if (field < 0) {
                field = 0
            }
        }

    /** 记录数 */
    var count:Int? = null
//        set(value) {
//            field = value
//            if (field < 1) {
//                field = 1
//            }
//        }

    /** 分组 */
    var group = SGroup()

    /**
     * 执行查询
     *
     * @return  返回查询到的记录
     */
    fun exec(): ArrayList<ENTITY> {
        return baseDao.execQuery(this)
    } // Fun exec()

    /**
     * 统计查询个数
     *
     * @return  返回查询到的记录个数
     */
    fun execCount(): Long {
        return baseDao.execCount(this)
    } // Fun execCount()

    /**
     * 执行查询，并返回第一条记录
     *
     * @return  返回查询到的第一条记录
     */
    fun entity(): ENTITY? {
        count = 1
        val list = exec()
        return if (list.size > 0) {
            list[0]
        } else {
            null
        }
    } // Fun entity()

    /**
     * 执行分页查询
     *
     * @param   pageNumber      第几页
     * @param   pageSize        每页记录数
     * @return  分页找到的数据
     */
    fun pageQuery(pageNumber: Int, pageSize: Int): SPageQueryInfo<ENTITY> {
        return baseDao.pageQuery(this, pageNumber, pageSize)
    } // Fun pageQuery()

    /**
     * 数据库表
     *
     * @param   name    表名
     * @return  查询构建器
     */
    fun table(name: String): SQueryBuilder<ENTITY> {
        tableName = name
        return this
    } // Fun tableName

    /**
     * 字段映射
     *
     * @param   fields      字段列表
     * @return  查询构建器
     */
    fun projection(fields: List<String>): SQueryBuilder<ENTITY> {
        for (name in fields) {
            fieldList.add(propertyName(name.trim()))
        }
        return this
    } // Fun projection()

    /**
     * 字段映射
     *
     * @param   fields      字段列表
     * @return  查询构建器
     */
    fun projection(fields: String): SQueryBuilder<ENTITY> {
        return projection(fields.split(","))
    } // Fun projection()

    /**
     * 添加过滤条件
     *
     * @param   filterList      过滤条件
     * @return  查询构建器
     */
    fun filter(vararg filterList: SFilter): SQueryBuilder<ENTITY> {
        for (item in filterList) {
            convertType(baseDao.entityClass, item)
            this.filterList.add(item)
        }
        return this
    } // Fun filter()

    /**
     * 添加过滤条件
     *
     * @param   filterList      过滤条件
     * @return  查询构建器
     */
    fun filter(vararg filterList: Pair<String, Any>): SQueryBuilder<ENTITY> {
        for (item in filterList) {
            val filter = SFilter.eq(item.first, item.second)
            convertType(baseDao.entityClass, filter)
            this.filterList.add(filter)
        }
        return this
    } // Fun filter()

    /**
     * 添加排序条件
     *
     * @param   name    排序属性名
     * @param   dir     排序方向
     * @return  查询构建器
     */
    fun order(name: String, dir: SDirection): SQueryBuilder<ENTITY> {
        orderArgs.add(SOrder(propertyName(name.trim()), dir))
        return this
    } // Fun order()

    /**
     * 添加排序条件
     *
     * @param   orderList   排序条件
     * @return  查询构建器
     */
    fun order(vararg orderList: SOrder): SQueryBuilder<ENTITY> {
        for (order in orderList) {
            order(order.name, order.direction)
        }
        return this
    } // Fun order()

    /**
     * 添加排序条件
     *
     * @param   orderStr    排序条件
     * @return  查询构建器
     */
    fun order(orderStr: String): SQueryBuilder<ENTITY> {
        val parser = SSyntaxParser(ByteArrayInputStream(orderStr.toByteArray()))
        val orderList = parser.parserOrders()
        order(*orderList.toTypedArray())
        return this
    } // Fun order()

    /**
     * 去除重复记录
     *
     * @return  查询构建器
     */
    fun distinct(): SQueryBuilder<ENTITY> {
        isDistinct = true
        return this
    } // Fun distinct()

    /**
     * 限定取得的记录数
     *
     * @param   count   取得的记录数
     * @return  查询构建器
     */
    fun limit(count: Int): SQueryBuilder<ENTITY> {
        limit(0, count)
        return this
    } // Fun limit()

    /**
     * 限定从指定位置取得的记录数
     *
     * @param   offset  指定位置(默认从0开始)
     * @param   count   取得的记录数
     * @return  查询构建器
     */
    fun limit(offset: Int, count: Int): SQueryBuilder<ENTITY> {
        this.offset = offset
        this.count = count
        return this
    } // Fun limit()

    /**
     * 添加分组
     *
     * @param   nameList    名称列表
     * @return  查询构建器
     */
    fun group(vararg nameList: String): SQueryBuilder<ENTITY> {
        for (name in nameList) {
            group.nameList.add(propertyName(name.trim()))
        }
        return this
    } // Fun group()

    /**
     * 是否统计记录个数
     */
    fun count(value: Boolean): SQueryBuilder<ENTITY>{
        group.count = value
        return this
    } // Fun count()
    /**
     * 统计最大数
     *
     * @param   nameList    统计的属性名称
     * @return  查询构建器
     */
    fun max(vararg nameList: String): SQueryBuilder<ENTITY> {
        for (name in nameList) {
            group.maxList.add(propertyName(name.trim()))
        }
        return this
    } // Fun max()


    /**
     * 统计最小数
     *
     * @param   nameList    统计的属性名称
     * @return  查询构建器
     */
    fun min(vararg nameList: String): SQueryBuilder<ENTITY> {
        for (name in nameList) {
            group.minList.add(propertyName(name.trim()))
        }
        return this
    } // Fun min()

    /**
     * 统计求和
     *
     * @param   nameList    统计的属性名称
     * @return  查询构建器
     */
    fun sum(vararg nameList: String): SQueryBuilder<ENTITY> {
        for (name in nameList) {
            group.sumList.add(propertyName(name.trim()))
        }
        return this
    } // Fun sum()

    /**
     * 统计求平均数
     *
     * @param   nameList    统计的属性名称
     * @return  查询构建器
     */
    fun avg(vararg nameList: String): SQueryBuilder<ENTITY> {
        for (name in nameList) {
            group.avgList.add(propertyName(name.trim()))
        }
        return this
    } // Fun avg()

    /**
     * 获得属性名
     *
     * @param   name    输入名称
     * @return  属性名
     */
    private fun propertyName(name: String): String {
        try {
            val clz = baseDao.entityClass
            return clz.getAllDeclaredField(name, ignoreCase = true, jsonProperty = true)?.name!!
        } catch (e: Exception) {
            e.printStackTrace()
            throw SFieldNotFoundException(name)
        }
    } // Fun propertyName()

    /**
     * 转换数据类型
     *
     * @param   entityClass 实体类类型
     * @param   filter      过滤条件
     */
    private fun convertType(entityClass: Class<*>, filter: SFilter) {
        if (filter.first is SFilter) {
            convertType(entityClass, filter.first as SFilter)
        }
        if (filter.second is SFilter) {
            convertType(entityClass, filter.second as SFilter)
        }

        if (filter.first is String) {
            val name = filter.first as String
            if (name.indexOf(".") < 0) {
                val field = entityClass.getAllDeclaredField(filter.first as String, ignoreCase = true, jsonProperty = true)
                        ?: return
                // 如果数据类型为日期型
                if (Date::class.java.isAssignableFrom(field.type)) {
                    val ft = SimpleDateFormat(SFormat.dateTime)
                    filter.second = ft.parse(filter.second as String)
                }
            }
        }
    } // Fun convertType()
} // Class SQueryBuilder
