/*
 * *********************************************************************************************************************
 *
 *          !!
 *        .F88X
 *        X8888Y
 *      .}888888N;
 *        i888888N;        .:!              .I$WI:
 *          R888888I      .'N88~            i8}+8Y&8"l8i$8>8W~'>W8}8]KW+8IIN"8&
 *          .R888888I    .;N8888~          .X8'  "8I.!,/8"  !%NY8`"8I8~~8>,88I
 *            +888888N;  .8888888Y                                  "&&8Y.}8,
 *            ./888888N;  .R888888Y        .'}~    .>}'.`+>  i}!    "i'  +/'  .'i~  !11,.:">,  .~]!  .i}i
 *              ~888888%:  .I888888l      .]88~`1/iY88Ii+1'.R$8$8]"888888888>  Y8$  W8E  X8E  W8888'188Il}Y88$*
 *              18888888    E8888881    .]W%8$`R8X'&8%++N8i,8N%N8+l8%`  .}8N:.R$RE%N88N%N$K$R  188,FE$8%~Y88I
 *            .E888888I  .i8888888'      .:$8I;88+`E8R:/8N,.>881.`$8E/1/]N8X.Y8N`"KF&&FK!'88*."88K./$88%RN888+~
 *            8888888I  .,N888888~        ~88i"8W,!N8*.I88.}888%F,i$88"F88"  888:E8X.>88!i88>`888*.}Fl1]*}1YKi'
 *          i888888N'      I888Y          ]88;/EX*IFKFK88X  K8R  .l8W  88Y  ~88}'88E&%8W.X8N``]88!.$8K  .:W8I
 *        .i888888N;        I8Y          .&8$  .X88!  i881.:%888>I88  ;88]  +88+.';;;;:.Y88X  18N.,88l  .+88/
 *      .:R888888I
 *      .&888888I                                          Copyright (c) 2009-2020.  博锐尚格科技股份有限公司
 *        ~8888'
 *        .!88~                                                                     All rights reserved.
 *
 * *********************************************************************************************************************
 */

package com.persagy.mybatis

import com.persagy.database.SFilter
import com.persagy.database.SGroup
import com.persagy.database.SQueryBuilder
import org.apache.ibatis.annotations.*

/**
 * 基础数据库操作影射器
 *
 * @author  庞利祥 <sybotan@126.com>
 */
interface IBaseMapper<ENTITY: Any> {
    /**
     * 根据条件删除
     *
     * @param   filterArgs      过滤条件
     * @param   whereArgs       whereArgs参数值，由SQlProvider填充
     */
    @DeleteProvider(type = SSqlProvider::class, method = "delete")
    fun delete(@Param("entityClass")entityClass: Class<*>,
               @Param("filterArgs")filterArgs: List<SFilter>,
               @Param("whereArgs")whereArgs: Map<String, Any>): Boolean

    /**
     * 删除所有记录
     */
    @DeleteProvider(type = SSqlProvider::class, method = "deleteAll")
    fun deleteAll(@Param("entityClass")entityClass: Class<*>): Boolean

    /**
     * 在数据库中插入实体
     *
     * @param   entity      插入数据库的实体
     * @return  插入是否成功
     */
    @InsertProvider(type = SSqlProvider::class, method = "insert")
    @Options(useGeneratedKeys = true, keyProperty = "entity._id")
    fun insert(@Param("entityClass")entityClass: Class<*>,
               @Param("entity")entity: Map<String, Any?>): Boolean

    /**
     * 在数据库中插入实体
     *
     * @param   entity      插入数据库的实体
     * @return  插入是否成功
     */
    @InsertProvider(type = SSqlProvider::class, method = "replace")
    fun replace(@Param("entityClass")entityClass: Class<*>,
                @Param("entity")entity: Map<String, Any?>,
                @Param("keyColumnList")keyColumnList: ArrayList<String>): Boolean

    /**
     * 根据条件更新
     *
     * @param   entity          更新值
     * @param   projection      字段映射
     * @param   filterArgs      过滤条件
     * @param   whereArgs       whereArgs参数值，由SQlProvider填充
     */
    @UpdateProvider(type = SSqlProvider::class, method = "update")
    fun update(@Param("entityClass")entityClass: Class<*>,
               @Param("entity")entity: Map<String, Any?>,
               @Param("projection")projection: ArrayList<String>?,
               @Param("filterArgs")filterArgs: List<SFilter>?,
               @Param("whereArgs")whereArgs: Map<String, Any>): Boolean

    /**
     * 执行查询操作
     *
     * @param   builder         查询查询构造器
     * @param   whereArgs       whereArgs参数值，由SQlProvider填充
     * @return  查询到的记录
     */
    @SelectProvider(type = SSqlProvider::class, method = "execQuery")
    fun execQuery(@Param("entityClass")entityClass: Class<*>,
                  @Param("builder")builder: SQueryBuilder<ENTITY>,
                  @Param("whereArgs")whereArgs: Map<String, Any>): ArrayList<Map<String, Any?>>

    /**
     * 执行查询操作
     *
     * @param   builder         查询查询构造器
     * @param   whereArgs       whereArgs参数值，由SQlProvider填充
     * @return  查询到的记录
     */
    @SelectProvider(type = SSqlProvider::class, method = "execCount")
    fun execCount(@Param("entityClass")entityClass: Class<*>,
                  @Param("builder")builder: SQueryBuilder<ENTITY>,
                  @Param("whereArgs")whereArgs: Map<String, Any>): ArrayList<Long>

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // 级联操作相关
    /**
     * 查询级联关系
     *
     * @param   table       表名
     * @param   id1         对象表ID字段名
     * @param   id2         级联对象ID字段名
     * @param   self        自身级联
     * @param   filter      附加过滤条件
     * @param   idList      对象ID列表
     * @return  关系列表
     */
    @SelectProvider(type = SSqlProvider::class, method = "relationQuery")
    fun relationQuery(@Param("table")table: String,
                      @Param("id1")id1: String,
                      @Param("id2")id2: String,
                      @Param("self")self: Boolean,
                      @Param("filter")filter: String,
                      @Param("idList")idList: ArrayList<Any>,
                      @Param("whereArgs")whereArgs: Map<String, Any>): ArrayList<HashMap<String, Any>>
} // Interface IBaseMapper
