/*
 * *********************************************************************************************************************
 *
 *          !!
 *        .F88X
 *        X8888Y
 *      .}888888N;
 *        i888888N;        .:!              .I$WI:
 *          R888888I      .'N88~            i8}+8Y&8"l8i$8>8W~'>W8}8]KW+8IIN"8&
 *          .R888888I    .;N8888~          .X8'  "8I.!,/8"  !%NY8`"8I8~~8>,88I
 *            +888888N;  .8888888Y                                  "&&8Y.}8,
 *            ./888888N;  .R888888Y        .'}~    .>}'.`+>  i}!    "i'  +/'  .'i~  !11,.:">,  .~]!  .i}i
 *              ~888888%:  .I888888l      .]88~`1/iY88Ii+1'.R$8$8]"888888888>  Y8$  W8E  X8E  W8888'188Il}Y88$*
 *              18888888    E8888881    .]W%8$`R8X'&8%++N8i,8N%N8+l8%`  .}8N:.R$RE%N88N%N$K$R  188,FE$8%~Y88I
 *            .E888888I  .i8888888'      .:$8I;88+`E8R:/8N,.>881.`$8E/1/]N8X.Y8N`"KF&&FK!'88*."88K./$88%RN888+~
 *            8888888I  .,N888888~        ~88i"8W,!N8*.I88.}888%F,i$88"F88"  888:E8X.>88!i88>`888*.}Fl1]*}1YKi'
 *          i888888N'      I888Y          ]88;/EX*IFKFK88X  K8R  .l8W  88Y  ~88}'88E&%8W.X8N``]88!.$8K  .:W8I
 *        .i888888N;        I8Y          .&8$  .X88!  i881.:%888>I88  ;88]  +88+.';;;;:.Y88X  18N.,88l  .+88/
 *      .:R888888I
 *      .&888888I                                          Copyright (c) 2009-2020.  博锐尚格科技股份有限公司
 *        ~8888'
 *        .!88~                                                                     All rights reserved.
 *
 * *********************************************************************************************************************
 */

package com.persagy.mybatis

import com.alibaba.fastjson.PropertyNamingStrategy
import com.persagy.base.extensions.*
import com.persagy.base.utils.SJsonUtil
import org.slf4j.LoggerFactory
import javax.persistence.Column

/**
 * 记录解析器
 *
 * @author  庞利祥(庞利祥)
 */
open class SRowParser<ENTITY: Any>(entityClazz: Class<ENTITY>) {
    companion object {
        /** 日志记录器 */
        private val logger = LoggerFactory.getLogger(SRowParser::class.java)
    }
    /** 实体类类型 */
    private val entityClass = entityClazz

    /**
     * 解析记录行
     *
     * @param   columns     字段映射
     * @return  实体类
     */
    open fun parseRow(columns: Map<String, Any?>): ENTITY {
        val entity = entityClass.getDeclaredConstructor().newInstance()
        for (field in entityClass.getAllDeclaredFields()) {
            val accessible = field.isAccessible    // canAccess(entity)
            field.isAccessible = true

            // 处理统计信息
            if (field.name == "statistics") {
                field.set(entity, groupStatistics(columns))
                continue
            }
            val anno = field.getAnnotation(Column::class.java) ?: continue

            val value = columns[anno.name] ?: continue
            try {
                when {
                    field.type.isBaseType() -> {
                        field.set(entity, value)
                    }
                    Map::class.java.isAssignableFrom(field.type) -> {
                        field.set(entity, SJsonUtil.fromJson(value.toString(), HashMap::class.java))
                    }
                    List::class.java.isAssignableFrom(field.type) -> {
                        try {
                            val genericType = field.genericType
                            val v1 = SJsonUtil.fromJson(value.toString(), genericType!!)
                            field.set(entity, v1)
                        } catch (e: Exception) {
                            // DO NOTHING
                        }
                    }
                    else -> {
                        // field.set(entity, SJsonUtil.fromJson(value.toString(), field.type))
                        field.set(entity, SJsonUtil.fromJson(value.toString(), field.type))
                    }
                }
            } catch (e: Exception) {
                e.printStackTrace()
            }
            field.isAccessible = accessible
        }

        return entity as ENTITY
    } // Fun parseRow()

    /**
     * 获得分组统计信息
     *
     * @param   columns     字段列表
     * @return  分组统计信息
     */
    private fun groupStatistics(columns: Map<String, Any?>): HashMap<String, Any?>? {
        var statistics: HashMap<String, Any?>?
        statistics = HashMap()
        for (col in columns) {
            if (!col.key.startsWith("_statistics_")) {
                continue
            }
            if (null == statistics) {
                statistics = HashMap()
            }

            when {
                col.key.startsWith("_statistics_count") ->
                    statistics[statisticName("Count", "")] = col.value
                col.key.startsWith("_statistics_avg_") ->
                    statistics[statisticName("Avg", col.key)] = col.value
                col.key.startsWith("_statistics_sum_") ->
                    statistics[statisticName("Sum", col.key)] = col.value
                col.key.startsWith("_statistics_max_") ->
                    statistics[statisticName("Max", col.key)] = col.value
                col.key.startsWith("_statistics_min_") ->
                    statistics[statisticName("Min", col.key)] = col.value
            }
        }

        return statistics
    } // Fun groupStatistics()

    /**
     * 统记字段名称
     *
     * @param   type        统计类型
     * @param   name        统计字段名
     * @return  字段名
     */
    private fun statisticName(type: String, name: String): String {
        val col = if (name.startsWith("_statistics_")) {
            name.substring(16)
        } else {
            ""
        }

        val str = "$col$type"
        return if (PropertyNamingStrategy.PascalCase == SJsonUtil.serializeConfig.propertyNamingStrategy) {
            str.firstUpperCase();
        } else {
            str.firstLowerCase();
        }
    } // Fun statisticName()
} // SRowParser